#include <stdio.h>
#include <string.h>
#include <list>

#include "parser.h"

//
// Expression context
//

// Expression parsing variables, same as in EXPRESSION, set in function Expression()
static PRECEDENCE_ITEM          *precedence_array = NULL;
static LIST_EXPR_ITEM           **precedence_array_cache = NULL;
static LIST_OPERAND             *list_cuo = NULL;
static LIST_OPERAND             *list_parentheses = NULL;
static LIST_OPERAND             *list_functions = NULL;

//
// Initializers
//

static CUO_ITEM     CUO_ITEM_init;
static EXPR_ITEM    EXPR_ITEM_init;
static EXPRESSION   EXPRESSION_init;

//
// Implementation
//

EXPR_ITEM *new_operand(TOKEN *token)
{
    EXPR_ITEM *expr_item;

    expr_item = new EXPR_ITEM(EXPR_ITEM_init);
    expr_item->t = EIT_OPERAND;

    expr_item->operand.t = OT_VALUE;
    expr_item->operand.value.t = OVT_VALUE_DOUBLE;
    expr_item->operand.value.d = token->num;

    return expr_item;
}

EXPR_ITEM *new_operand(LIST_CUO_ITEM *list_cuo_item)
{
    EXPR_ITEM *expr_item;

    expr_item = new EXPR_ITEM(EXPR_ITEM_init);
    expr_item->t = EIT_OPERAND;

    expr_item->operand.t = OT_CUO;
    expr_item->operand.cuo = list_cuo_item;

    return expr_item;
}

EXPR_ITEM *new_operand(LIST_EXPR_ITEM *list_expr_item)
{
    EXPR_ITEM *expr_item;

    expr_item = new EXPR_ITEM(EXPR_ITEM_init);
    expr_item->t = EIT_OPERAND;

    expr_item->operand.t = OT_LIST;
    expr_item->operand.list = list_expr_item;

    return expr_item;
}

EXPR_ITEM *new_operator(TOKEN *token)
{
    EXPR_ITEM *expr_item;

    expr_item = new EXPR_ITEM(EXPR_ITEM_init);
    expr_item->t = EIT_OPERATOR;

    expr_item->bop = token->op;

    return expr_item;
}

CUO_ITEM *new_cuo_item(TOKEN_OPERATOR_TYPE op)
{
    CUO_ITEM *cuo_item;

    cuo_item = new CUO_ITEM(CUO_ITEM_init);
    cuo_item->t = CIT_TOKEN_OPERATOR;

    cuo_item->uop = op;

    return cuo_item;
}

CUO_ITEM *new_cuo_item(double d)
{
    CUO_ITEM *cuo_item;

    cuo_item = new CUO_ITEM(CUO_ITEM_init);
    cuo_item->t = CIT_OPERAND_VALUE;

    cuo_item->value.t = OVT_VALUE_DOUBLE;
    cuo_item->value.d = d;

    return cuo_item;
}

int get_operator_precedence(TOKEN_OPERATOR_TYPE op)
{
    switch (op)
    {
        case TOT_LTHAN:
        case TOT_GTHAN:
            return OP_0;
        case TOT_MINUS:
        case TOT_PLUS:
            return OP_1;
        case TOT_MUL:
        case TOT_DIV:
            return OP_2;
        default:
            fprintf(stderr, "get_operator_precedence(): Unhandled operator\n");
            break;
    }
    return 0;
}

LIST_EXPR_ITEM * first_list_from_cache_with_higher_precedence(int higher_than_precedence)
{
    for (int i = higher_than_precedence + 1; i < OP_MAX; i++) {
        if (precedence_array_cache[i] != NULL)
            return precedence_array_cache[i];
    }
    return NULL;
}

// Function iterates precedence_array_cache starting from higher_than_precedence, it does:
// 1. Check if last item in list is pointer to another list with value NULL, if yes, replace
// NULL value with pointer to first list with higher precedence.
// 2. Save list from cache to precedence_array and set pointer in precedence_array_cache to NULL.
// If input parameter higher_than_precedence is < 0 then:
//     Step (1) and step (2) apply on all lists in cache
// else:
//     Step (1) apply on list in cache with precedence >= higher_than_precedence
//     Step (2) apply on list in cache with precedence > higher_than_precedence
void save_and_clear_cache(int higher_than_precedence)
{
    LIST_EXPR_ITEM  *list;
    EXPR_ITEM       *expr_item;
    bool            all_items = false;

    if (higher_than_precedence < 0) {
        higher_than_precedence = 0;
        all_items = true;
    }

    for (int i = higher_than_precedence; i < OP_MAX; i++) {
        list = precedence_array_cache[i];
        if (list == NULL)
            continue;
        expr_item = list->back();
        // If latest item in list is NULL, then replace it with pointer to first list with higher precedence
        if (expr_item->t != EIT_OPERAND) {
            fprintf(stderr, "save_and_clear_cache(): error\n");
            return;
        }
        if ((expr_item->operand.t == OT_LIST) && (expr_item->operand.list == NULL)) {
            expr_item->operand.list = first_list_from_cache_with_higher_precedence(i);
            if (expr_item->operand.list == NULL) {
                fprintf(stderr, "save_and_clear_cache(): error\n");
                return;
            }
        }
        if (!all_items && (i <= higher_than_precedence))
            continue;
        precedence_array[i].push_back(list);
        precedence_array_cache[i] = NULL;
    }
}

LIST_CUO_ITEM *parse_cuo(TOKEN *token)
{
    LIST_CUO_ITEM   *list_cuo_item = new LIST_CUO_ITEM();

    // Get token until Operand value
    while (1) {
        // Test if token is in allowed unary operators
        if (token->t == TT_OPERATOR)
            switch (token->op) {
                case TOT_MINUS:
                case TOT_PLUS:
                    break;
                default:
                    fprintf(stderr, "parse_cuo(): unsupported unary operator\n");
                    return NULL;
            }
        else {
            fprintf(stderr, "parse_cuo(): unexpected token type\n");
            return NULL;
        }

        // Add token operator to list
        list_cuo_item->push_back(new_cuo_item(token->op));

        token_Get(token);
        if (token == NULL) {
            fprintf(stderr, "parse_cuo(): error\n");
            return NULL;
        }

        // If token is value, add it to list and return list_cuo_item
        if (token->t == TT_VAL_NUM) {
            list_cuo_item->push_back(new_cuo_item(token->num));
            break;
        }
    } // while (1)

    return list_cuo_item;
}

// function called from Expression()
int _Expression(void)
{
    TOKEN           _token;
    TOKEN           *token = &_token;

    EXPR_ITEM       *expr_item_operand;
    EXPR_ITEM       *expr_item_operator;

    LIST_EXPR_ITEM  *list_current = NULL;
    LIST_EXPR_ITEM  *list_first;

    int             precedence_prev = OP_NONE;
    int             precedence;

    while (1) {
        token_Get(token);
        if (token == NULL)
            break;

        // If first token is operator, parse chain of unary operators with operand value
        if (token->t == TT_OPERATOR) {
            expr_item_operand = new_operand(parse_cuo(token));
            // save CUO in list_cuo for printing
            list_cuo->push_back(&(expr_item_operand->operand));
        } else
            expr_item_operand = new_operand(token);

        token_Get(token);

        // If operand is only item in expression:
        // 1. Add operand to precedence array
        // 2. Parsing of expression is done
        // If operand is last item in expression:
        // 1. Add operand to current list
        // 2. Save and clear cache for all items in cache
        // 3. Parsing of expression is done
        if (token == NULL) {
            if (list_current == NULL) {
                // Operand is first item in expression
                list_current = new LIST_EXPR_ITEM();
                list_current->push_back(expr_item_operand);
                precedence_array[OP_0].push_back(list_current);
                break;
            }
            // Operand is last item in expression
            list_current->push_back(expr_item_operand);
            save_and_clear_cache(-1);
            break;
        } else
            expr_item_operator = new_operator(token);

        precedence = get_operator_precedence(token->op);

        // Init precedence_prev with first operator
        if (precedence_prev == OP_NONE)
            precedence_prev = precedence;

        // Init first list
        if (list_current == NULL) {
            list_current = new LIST_EXPR_ITEM();
            list_first = list_current;
            precedence_array_cache[precedence] = list_current;
        }

        // If operator has same precedence as previous operator:
        // 1. add operand and operator to current list
        if (precedence_prev == precedence) {
        }

        // If operator has lower precedence than previous operator:
        // 1. Save operand to current list
        // 2. Allocate new operand containing pointer to list:
        //    - from cache get first list with higher precedence than current precedence
        //    - if list from cache is NULL, get first list in expression
        // 3. Save and clear cache
        // 4. Get new current list:
        //    - list for current precedence from cache 
        //    - if list from cache is NULL, allocate new list and store it in cache
        // 5. If current list was not allocated in step (4), set operand to NULL
        // 6. Add operand and operator to current list
        if (precedence_prev > precedence) {

            list_current->push_back(expr_item_operand);

            list_current = first_list_from_cache_with_higher_precedence(precedence);
            if (list_current == NULL)
                list_current = list_first;
            expr_item_operand = new_operand(list_current);

            save_and_clear_cache(precedence);

            list_current = precedence_array_cache[precedence];
            if (list_current == NULL) {
                list_current = new LIST_EXPR_ITEM();
                precedence_array_cache[precedence] = list_current;
            } else
                expr_item_operand = NULL;
        }

        // If operator has higher precedence than previous operator:
        // 1. Create new operand with NULL pointer to list as marker
        // 2. Save operand from step (1) to current list
        // 3. Get new current list:
        //    - list for current precedence from cache 
        //    - if list from cache is NULL, allocate new list and store it in cache
        // 4. Add operand and operator to current list
        if (precedence_prev < precedence) {

            list_current->push_back( new_operand((LIST_EXPR_ITEM *)NULL) );

            list_current = precedence_array_cache[precedence];
            if (list_current == NULL) {
                list_current = new LIST_EXPR_ITEM();
                precedence_array_cache[precedence] = list_current;
            }
        }

        // Add operand and operator to current list
        if (expr_item_operand != NULL)
            list_current->push_back(expr_item_operand);
        list_current->push_back(expr_item_operator);

        // Save previous priority
        precedence_prev = precedence;
    }

    return 0;
}

EXPRESSION * Expression(void)
{
    EXPRESSION *expression = new EXPRESSION(EXPRESSION_init);

    // Set parser variables
    precedence_array = expression->precedence_array;
    precedence_array_cache = expression->precedence_array_cache;
    list_cuo = & expression->list_cuo;
    list_parentheses = & expression->list_parentheses;
    list_functions = & expression->list_functions;

    // Call expression
    if (0 != _Expression())
        return NULL;

    return expression;
}
